import logbook

from flask import Blueprint
from .base import SubsystemBase

_logger = logbook.Logger(__name__)

class FlaskBlueprintSubsystem(SubsystemBase):

    NAME = 'blueprint'

    def configure_grain(self, grain, app):
        _logger.trace('Found blueprint: {}', grain)
        main = grain.load()
        name = None
        for name in dir(main):
            if name.startswith('_'):
                continue
            blueprint = getattr(main, name)
            if isinstance(blueprint, Blueprint):
                break
        else:
            raise RuntimeError('Could not find any blueprint in {}'.format(grain.path))
        url_prefix = grain.config.get('mountpoint', '/')
        _logger.trace('registering {}:{} under {}', grain, name, url_prefix)
        app.register_blueprint(blueprint, url_prefix=url_prefix)



import os

import click
import logbook
import yaml

from ..bootstrapping import (ensure_project_bootstrapped,
                             get_virtualenv_binary_path)
from ..project import get_project

_logger = logbook.Logger(__name__)


@click.command()
def develop():
    ensure_project_bootstrapped()
    tmuxp = get_virtualenv_binary_path('tmuxp')
    tmux_config_filename = os.path.abspath('.cob/_frontend.yml')
    with open(tmux_config_filename, 'w') as f:
        yaml.dump(_get_tmux_config(), f)
    os.execve(tmuxp, [tmuxp, 'load', tmux_config_filename], {**os.environ})


def _get_tmux_config():

    windows = [
        {
            'window_name': 'flask',
            'layout': 'even-horizontal',
            'panes': [
                'source .cob/env/bin/activate && cob testserver',
            ]
        },
    ]

    for subsystem in get_project().subsystems:
        subsystem.configure_tmux_window(windows)

    return {
        'session_name': 'cob-{}'.format(get_project().name),
        'windows': windows,
    }



import functools

from ..app import build_app


def appcontext_command(func):
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        with build_app().app_context():
            return func(*args, **kwargs)

    return new_func



import os
from contextlib import contextmanager

import jinja2
import logbook


_logger = logbook.Logger(__name__)


@contextmanager
def chdir_context(path):
    path = str(path)
    oldpath = os.path.abspath('.')
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(oldpath)


def write_template(fileobj, name, args):
    with open(os.path.join(os.path.dirname(__file__), '_templates', name)) as f:
        template = jinja2.Template(f.read())
    fileobj.write(template.render(**args))



import os

import emport

_SUBSYSTEM_BY_NAME = {}

class SubsystemMeta(type):

    def __new__(mcs, classname, bases, classdict):
        returned = type.__new__(mcs, classname, bases, classdict)
        if 'NAME' in classdict:
            _SUBSYSTEM_BY_NAME[classdict['NAME']] = returned
        return returned


class SubsystemBase(metaclass=SubsystemMeta):

    SUBSYSTEM_BY_NAME = _SUBSYSTEM_BY_NAME

    def __init__(self, manager):
        super(SubsystemBase, self).__init__()
        self.manager = manager
        self.project = self.manager.project
        self.grains = []

    def add_grain(self, path, config):
        self.grains.append(LoadedGrain(self, path, config))

    def activate(self, flask_app):
        pass

    def configure_app(self, flask_app):
        for grain in self.grains:
            self.configure_grain(grain, flask_app)

    def configure_grain(self, grain, flask_app):
        raise NotImplementedError() # pragma: no cover

    def configure_tmux_window(self, windows):
        pass


class LoadedGrain(object):

    def __init__(self, subsystem, path, config):
        super(LoadedGrain, self).__init__()
        self.subsystem = subsystem
        self.path = path
        self.config = config

    def load(self):
        if os.path.isfile(self.path):
            return emport.import_file(self.path)
        main = self.config.get('main', 'main')
        if not main.endswith('.py'):
            main += '.py'
        return emport.import_file(os.path.join(self.path, main))

    def load_python_symbol_by_name(self, symbol):
        filename, symbol = symbol.rsplit(':', 1)
        grain = self.load_python_module_by_name(filename)
        return getattr(grain, symbol)

    def load_python_module_by_name(self, rel_filename):
        assert not os.path.isabs(rel_filename)
        rel_filename = os.path.join(self.path, rel_filename)
        if not os.path.isfile(rel_filename) and not rel_filename.endswith('.py'):
            rel_filename += '.py'
        if not os.path.isfile(rel_filename):
            raise RuntimeError('File does not exist: {!r}'.format(rel_filename))
        module = emport.import_file(rel_filename)
        return module

    def __repr__(self):
        return self.path



import logbook

from .base import SubsystemBase

from flask import Blueprint

_logger = logbook.Logger(__name__)

class ViewsSubsystem(SubsystemBase):

    NAME = 'views'

    def configure_grain(self, grain, app):
        _logger.trace('Found views grain: {}', grain)
        main = grain.load()
        blueprint = Blueprint(main.__name__, main.__name__, url_prefix=grain.config['mountpoint'])
        for deferred_route in getattr(main, '__cob_routes__', []):
            deferred_route.register(blueprint)
        app.register_blueprint(blueprint)



import os
from uuid import uuid4

import logbook

import pytest

from .project import Project
from cob.bootstrapping import _PREVENT_REENTRY_ENV_VAR
from cob.utils.develop import _COB_DEVELOP_MODE


@pytest.fixture(scope='session', autouse=True)
def setup_logging():
    logbook.StderrHandler(level=logbook.TRACE).push_application()


@pytest.fixture(scope='session', autouse=True)
def prevent_bootstrapping():
    os.environ[_PREVENT_REENTRY_ENV_VAR] = 'true'
    os.environ[_COB_DEVELOP_MODE] = 'true'


@pytest.fixture
def project_name():
    return 'proj{}'.format(str(uuid4()).replace('-', '_'))


@pytest.fixture
def empty_project(tmpdir, project_name):
    returned = Project(tmpdir, project_name)
    returned.generate_project()
    return returned



import os

import logbook

from .base import SubsystemBase

_logger = logbook.Logger(__name__)



class StaticSubsystem(SubsystemBase):

    NAME = 'static'

    def configure_grain(self, grain, app): # pylint: disable=unused-argument
        cfg = grain.config
        root = os.path.join(grain.path, cfg['root'])
        self.project.add_static_location(cfg['mountpoint'], root)



import os
import re
import shutil
import socket
import subprocess
import sys
import tempfile
import time
from contextlib import contextmanager

import logbook
from jinja2 import Environment as TemplateEnvironment
from jinja2 import FileSystemLoader

import requests
from urlobject import URLObject

_logger = logbook.Logger(__name__)

_DEFAULT_PORT = 6789

_PROJS_ROOT = os.path.join(os.path.dirname(__file__), '_projects')


template_env = TemplateEnvironment(loader=FileSystemLoader(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), '_templates')))


class Project(object):

    def __init__(self, name):
        super(Project, self).__init__()
        self._name = name
        self.tempdir = tempfile.mkdtemp()
        self.projdir = os.path.join(self.tempdir, 'proj')
        shutil.copytree(os.path.join(_PROJS_ROOT, self._name), self.projdir)
        self.logfile_name = os.path.join(self.tempdir, 'testserver.log')

    def cmd(self, cmd):
        return subprocess.check_call(cmd, shell=True, cwd=self.projdir)

    def on(self, path):
        return ProjectPath(self, path)

    @contextmanager
    def server_context(self):

        _logger.debug('Server log is going to {}...', self.logfile_name)

        with open(self.logfile_name, 'a') as logfile:

            self._run_cob(['bootstrap'], logfile).wait()
            with self._end_killing(self._run_cob(['testserver', '-p', '0'], logfile)) as p:
                port = self._parse_port(logfile)
                self._wait_for_server(process=p, port=port)
                yield URLObject('http://127.0.0.1:{}'.format(port))

    def _run_cob(self, argv, logfile):
        return subprocess.Popen(
            ' '.join([sys.executable, '-m', 'cob.cli.main', '-vvvvv', *argv]),
            cwd=self.projdir,
            shell=True,
            stdout=logfile,
            stderr=subprocess.STDOUT,
        )

    @contextmanager
    def _end_killing(self, p):
        try:
            yield p
        finally:
            _logger.debug('Killing server')
            if p.poll() is None:
                p.terminate()
            p.wait()

    def _parse_port(self, logfile, timeout_seconds=5):
        end_time = time.time() + timeout_seconds
        while time.time() < end_time:
            with open(logfile.name, 'r') as f:
                for line in f:
                    match = re.search(
                        r'Running on http://127\.0\.0\.1:(\d+)', line)
                    if match:
                        return int(match.group(1))
            time.sleep(0.1)
        with open(logfile.name, 'r') as f:
            error_msg = 'Could not parse port. It is very likely that cob has failed or encountered an exception.\nOutput was:\n {})'.format(f.read())
            raise RuntimeError(error_msg)

    def _wait_for_server(self, port, timeout_seconds=2, process=None):
        end_time = time.time() + timeout_seconds
        while time.time() < end_time:
            s = socket.socket()
            try:
                s.connect(('127.0.0.1', port))
            except socket.error:
                if process is not None and process.poll() is not None:
                    raise RuntimeError('Process exited!')
                time.sleep(0.05)
                continue
            else:
                break
        else:
            raise RuntimeError('Could not connect')

class ProjectPath(object):

    def __init__(self, project, path):
        super(ProjectPath, self).__init__()
        self.project = project
        self.path = path

    def returns(self, code_or_string):
        resp = self._request()
        if isinstance(code_or_string, int):
            assert resp.status_code == code_or_string
        else:
            assert resp.text == code_or_string
        return True

    def returns_json(self, value):
        assert self._request().json() == value
        return True

    def _request(self):
        with self.project.server_context() as url:
            return requests.get(url.add_path(self.path))



import pytest
import requests

from .project import Project

@pytest.mark.parametrize('projname', ['blueprints1', 'blueprints2', 'blueprints3'])
def test_blueprints(projname):
    with Project(projname).server_context() as url:
        for name in ('bp1', 'bp2'):
            assert requests.get(url.add_path(name).add_path('test')).content.decode(
                'utf-8') == 'this is {}'.format(name)



import os
import click
import logbook
import subprocess

from jinja2 import Template
from pkg_resources import resource_string
from ..utils.develop import is_develop, cob_branch
from ..project import get_project
from cob.__version__ import __version__


_logger = logbook.Logger(__name__)
_CUSTOM_DOCKERFILE = "custom.docker"


def _get_user_steps():
    if not os.path.isfile(_CUSTOM_DOCKERFILE):
        return ''

    with open(_CUSTOM_DOCKERFILE) as f:
        return f.read()


def _generate_cob_installation():
    if is_develop():
        return "ENV COB_DEVELOP=1\nRUN cd /opt && git clone -b {} https://github.com/getweber/cob && pip install -e cob --no-cache-dir".format(cob_branch())
    else:
        return "RUN pip install cob=={}".format(__version__)


@click.group()
def docker():
    pass


@docker.command()
def generate():
    dockerfile_template = Template(resource_string("cob", "Dockerfile.j2").decode("UTF-8"))
    with open(".Dockerfile", "w") as f:
        f.write(dockerfile_template.render(
            install_cob=_generate_cob_installation(),
            user_steps=_get_user_steps()))


@docker.command()
@click.option('--sudo', is_flag=True, help="Run docker build with sudo")
@click.option('--extra-build-args', '-e', default="", help="Arguments to pass to docker build")
def build(sudo, extra_build_args):
    project = get_project()
    generate.callback()

    cmd = "{}docker build -t {} -f .Dockerfile {} .".format(
        "sudo " if sudo else "",
        project.name,
        extra_build_args)

    subprocess.check_call(cmd, shell=True)



import os
import subprocess
import sys

import click
import logbook
import venv
import yaml

from .defs import COB_CONFIG_FILE_NAME
from .utils.develop import is_develop, cob_root
from .project import get_project

_logger = logbook.Logger(__name__)

_PREVENT_REENTRY_ENV_VAR = 'COB_NO_REENTRY'
_COB_REFRESH_ENV = 'COB_REFRESH_ENV'
_VIRTUALENV_PATH = '.cob/env'
_INSTALLED_DEPS = '.cob/_installed_deps.yml'

def ensure_project_bootstrapped():
    if not os.path.isfile(COB_CONFIG_FILE_NAME):
        _logger.trace('Project is not a cob project')
        return
    if _PREVENT_REENTRY_ENV_VAR in os.environ:
        _logger.trace('{} found in environ. Not reentering.', _PREVENT_REENTRY_ENV_VAR)
        return
    _ensure_virtualenv()
    _reenter()

def get_virtualenv_binary_path(name):
    return os.path.join(_VIRTUALENV_PATH, 'bin', name)

def _ensure_virtualenv():
    if not _needs_refresh():
        _logger.trace('Virtualenv already seems bootstrapped. Skipping...')
        return
    _logger.trace('Creating virtualenv in {}', _VIRTUALENV_PATH)
    venv_parent_dir = os.path.dirname(os.path.abspath(_VIRTUALENV_PATH))
    if not os.path.isdir(venv_parent_dir):
        os.makedirs(venv_parent_dir)
    venv.create(_VIRTUALENV_PATH)
    subprocess.check_call([os.path.join(_VIRTUALENV_PATH, 'bin', 'python'), '-m', 'ensurepip'])
    if is_develop():
        _logger.trace('Using development version of cob')
        _virtualenv_pip_install(['-e', cob_root()])
    else:
        _logger.trace('Installing cob form Pypi')
        _virtualenv_pip_install(['-U', 'cob'])

    deps = sorted(get_project().get_deps())
    _virtualenv_pip_install(['-U', *deps])
    with open(_INSTALLED_DEPS, 'w') as f:
        yaml.dump(deps, f)

def _needs_refresh():
    if _COB_REFRESH_ENV in os.environ:
        click.echo(click.style('Virtualenv refresh forced. This might take a while...', fg='magenta'))
        return True
    if not os.path.exists(os.path.join(_VIRTUALENV_PATH, 'bin', 'python')):
        click.echo(click.style('Creating project environment. This might take a while...', fg='magenta'))
        return True
    if _get_installed_deps() != get_project().get_deps():
        click.echo(click.style('Dependencies have changes - refreshing virtualenv. This might take a while...', fg='magenta'))
        return True
    return False

def _get_installed_deps():
    if not os.path.isfile(_INSTALLED_DEPS):
        return set()
    with open(_INSTALLED_DEPS) as f:
        return set(yaml.load(f.read()))

def _virtualenv_pip_install(argv):
    _logger.trace('Installing cob in virtualenv...')
    subprocess.check_call([os.path.join(_VIRTUALENV_PATH, 'bin', 'python'), '-m', 'pip', 'install', *argv])

def _reenter():
    argv = sys.argv[:]
    argv[:1] = [os.path.abspath(os.path.join(_VIRTUALENV_PATH, 'bin', 'python')), '-m', 'cob.cli.main']
    _logger.trace('Running in {}: {}...', _VIRTUALENV_PATH, argv)
    os.execve(argv[0], argv, {_PREVENT_REENTRY_ENV_VAR: 'true', **os.environ})

def _which(bin):
    for directory in os.environ['PATH'].split(':'):
        full_path = os.path.join(directory, bin)
        if os.path.isfile(full_path):
            return full_path

    raise ValueError('Could not find a python interpreter named {}'.format(bin))



import os

from .base import SubsystemBase

class FrontendSubsystem(SubsystemBase): # pylint: disable=abstract-method
    pass

class EmberSubsystem(FrontendSubsystem):

    NAME = 'frontend-ember'

    def configure_grain(self, grain, flask_app): # pylint: disable=unused-argument
        self.project.add_static_location('/assets', os.path.join(grain.path, 'dist/assets'))
        self.project.add_static_file_alias('/', os.path.join(grain.path, 'dist/index.html'))

    def configure_tmux_window(self, windows):
        for grain in self.grains:
            windows.append({
                'window_name': 'frontend({})'.format(os.path.basename(grain.path)),
                'panes': [
                    'cd "{}" && ember build --watch'.format(grain.path),
                ],
            })



class ProjectContext(object):

    __slots__ = ['db']

    def __init__(self):
        super(ProjectContext, self).__init__()
        self.db = None


class _ContextStack(object):

    def __init__(self):
        super(_ContextStack, self).__init__()
        self._stack = [ProjectContext()]

    def __getattr__(self, attr):
        if not self._stack:
            raise AttributeError(attr)
        return getattr(self._stack[-1], attr)

    def __dir__(self):
        if self._stack:
            return dir(self._stack[-1])
        return []

    def __setattr__(self, attr, value):
        if attr.startswith("_"):
            return super(_ContextStack, self).__setattr__(attr, value)
        setattr(self._stack[-1], attr, value)

    def push(self, ctx):
        self._stack.append(ctx)
        return ctx

    def pop(self):
        assert len(self._stack) != 0
        if len(self._stack) == 1:
            raise RuntimeError("No more contexts to pop")
        return self._stack.pop(-1)

context = _ContextStack()

class ContextAttributeProxy(object):

    def __init__(self, name):
        super(ContextAttributeProxy, self).__init__()
        self._proxy__name = name

    @property
    def _obj(self):
        return getattr(context, self._proxy__name)

    def __getattr__(self, attr):
        return getattr(self._obj, attr)

    def __setattr__(self, attr, value):
        if attr == "_proxy__name":
            return super(ContextAttributeProxy, self).__setattr__(attr, value)
        setattr(self._obj, attr, value)

    def __eq__(self, other):
        return self._obj == other

    def __ne__(self, other):
        return self._obj != other

    def __call__(self, *args, **kwargs):
        return self._obj(*args, **kwargs)  # pylint: disable=not-callable

    def __repr__(self):
        return repr(self._obj)

    def __dir__(self):
        return dir(self._obj)

    __members__ = __dir__

    def __str__(self):
        return str(self._obj)


db = ContextAttributeProxy("db")



import sys
import copy


def route(*args, **kwargs):
    def callable(func):
        calling_module_globals = sys._getframe(1).f_globals # pylint: disable=protected-access
        routes = calling_module_globals.setdefault('__cob_routes__', [])
        routes.append(DeferredRoute(args, kwargs, func))
        return func
    return callable


class DeferredRoute(object):

    def __init__(self, args, kwargs, func):
        super(DeferredRoute, self).__init__()
        self.args = copy.deepcopy(args)
        self.kwargs = copy.deepcopy(kwargs)
        self.func = func

    def register(self, parent):
        parent.route(*self.args, **self.kwargs)(self.func)



import os
import subprocess
import sys

import pytest


def test_adding_deps(tmpdir):
    with pytest.raises(ImportError):
        import sentinels

    projdir = tmpdir.join('proj')
    yaml = projdir.join('.cob-project.yml')
    python = str(projdir.join('.cob/env/bin/python'))


    with yaml.open('a', ensure=True) as f:
        print('name: testproj', file=f)
    _cob_on(projdir, 'bootstrap')

    assert os.path.exists(python)

    assert subprocess.call([python, '-c', 'import sentinels']) == 1

    with yaml.open('a') as f:
        print('deps:', file=f)
        print('  - sentinels', file=f)

    _cob_on(projdir, 'bootstrap')
    assert subprocess.call([python, '-c', 'import sentinels']) == 0


def _cob_on(cwd, cmd):
    x = os.environ.pop('COB_NO_REENTRY')
    try:
        subprocess.check_call([sys.executable, '-m', 'cob.cli.main', '-vvvvv', str(cmd)], cwd=str(cwd))
    finally:
        os.environ['COB_NO_REENTRY'] = x



import requests

from .project import Project


def test_empty_project():
    assert Project('empty').on('/').returns(requests.codes.not_found)

def test_simple_views():
    assert Project('simple').on('/').returns('hey')



import click
import logbook

from .utils import appcontext_command

import flask_migrate

_logger = logbook.Logger(__name__)


@click.group()
def migrate():
    pass


@migrate.command()
@appcontext_command
def init():
    flask_migrate.init('migrations', False)


@migrate.command()
@click.option('-m', '--message', default=None)
@appcontext_command
def revision(message):
    flask_migrate.upgrade()
    flask_migrate.revision(autogenerate=True, message=message)

@migrate.command()
@appcontext_command
def up():
    flask_migrate.upgrade()

@migrate.command()
@appcontext_command
def down():
    flask_migrate.upgrade()



import click
import logbook

from cob.ctx import context

from .utils import appcontext_command

_logger = logbook.Logger(__name__)


@click.group()
def db():
    pass

@db.command()
@appcontext_command
def createall():
    context.db.create_all()



import logbook

from .project import get_project

_logger = logbook.Logger(__name__)


_cached_app = None

def build_app(*, use_cached=False):
    from flask import Flask

    global _cached_app  # pylint: disable=global-statement
    if use_cached and _cached_app is not None:
        return _cached_app


    proj = get_project()
    _logger.debug('Starting app {.name}...', proj)
    _cached_app = Flask(get_project().name, static_folder=None, template_folder=None)
    proj.configure_app(_cached_app)
    _logger.trace('URL map: {}', _cached_app.url_map)
    return _cached_app



import os

import yaml
import logbook

from .defs import COB_CONFIG_FILE_NAME
from .subsystems.manager import SubsystemsManager

from flask.helpers import send_from_directory
from flask import abort

_projet = None

_logger = logbook.Logger(__name__)


class Project(object):

    def __init__(self):
        super(Project, self).__init__()
        self.root = os.path.abspath('.')
        with open(os.path.join(self.root, COB_CONFIG_FILE_NAME)) as f:
            self.config = yaml.load(f)
        self.name = self.config.get('name', os.path.basename(self.root))
        self.subsystems = SubsystemsManager(self)

        self.static_locations = {}
        self.static_aliases = {}

    def get_deps(self):
        return set(self.config.get('deps') or ())

    def configure_app(self, app):
        app.config.update(self.config.get('flask_config', {}))
        self.subsystems.configure_app(app)
        self._configure_static_locations(app)

    def add_static_location(self, url_path, fs_path):
        self.static_locations.setdefault(url_path, []).append(os.path.abspath(fs_path))

    def add_static_file_alias(self, url_path, fs_path):
        assert url_path not in self.static_aliases
        self.static_aliases[url_path] = os.path.abspath(fs_path)

    def _configure_static_locations(self, flask_app):
        for url_path, fs_paths in self.static_locations.items():
            if not url_path.endswith('/'):
                url_path += '/'
            flask_app.route(url_path + '<path:filename>',
                            defaults={'search_locations': fs_paths})(_static_view)

        for url_path, fs_path in self.static_aliases.items():
            flask_app.route(url_path, defaults={'path': fs_path})(
                _static_alias_view)


def _static_view(filename, search_locations):
    for location in search_locations:
        if not location.endswith('/'):
            location += '/'
        _logger.trace('looking for {!r} in {!r}...', filename, location)
        p = os.path.join(location, filename)
        if os.path.exists(p) and os.path.abspath(p).startswith(location):
            _logger.trace('Found. Serving')
            return send_from_directory(location, filename)
    abort(404)


def _static_alias_view(path):
    return send_from_directory(os.path.dirname(path), os.path.basename(path))


_project = None


def get_project():
    global _project  # pylint: disable=global-statement
    if _project is None:
        _project = Project()

    return _project



import logbook

from .base import SubsystemBase

_logger = logbook.Logger(__name__)

class FlaskBlueprintSubsystem(SubsystemBase):

    NAME = 'templates'

    def configure_grain(self, grain, app):
        if app.template_folder is not None:
            raise RuntimeError('Multiple template dirs not supported yet')

        app.template_folder = grain.path



from io import StringIO

import pytest

from cob.utils.parsing import parse_front_matter

@pytest.mark.parametrize('line', [
    "# cob: x=value1 y=value2",
    "#cob:   x=value1    y=value2",
])
def test_front_matter_parser_oneline(line):
    assert _parse(line) == {'x': 'value1', 'y': 'value2'}

@pytest.mark.parametrize('line', [
    '# cob: x',
    '#',
])
def test_front_matter_parser_oneline_wrong(line):
    assert _parse(line) is None

def test_front_matter_yaml():
    yaml = _parse(
"""# cob-yaml:
# type: x
# value: y
# ---
""")
    assert yaml == {'type': 'x', 'value': 'y'}


def _parse(s):
    f = StringIO(s)
    return parse_front_matter(f)



import os

import yaml
import logbook

from .base import SubsystemBase
from ..utils.parsing import parse_front_matter

_logger = logbook.Logger(__name__)


class SubsystemsManager(object):

    def __init__(self, project):
        super(SubsystemsManager, self).__init__()
        self.project = project
        self._subsystems = {}
        self._load_project_subsystems()

    def _load_project_subsystems(self):
        roots = [self.project.root]
        while roots:
            root = roots.pop()
            for name in os.listdir(root):
                _logger.trace('Examining {}...', name)
                path = os.path.join(root, name)
                config = self._try_get_config(path)
                if config is None:
                    _logger.trace('{} does not seem to be a cob grain. Skipping...', name)
                    continue
                _logger.trace(
                    'Detected grain in {} (subsystem: {[type]}', name, config)
                if config['type'] == 'bundle':
                    _logger.trace('Will traverse into bundle {}', path)
                    roots.append(path)
                    continue
                subsystem_cls = self._get_subsystem_by_grain_type(config['type'])
                subsystem = self._subsystems.get(subsystem_cls.NAME)
                if subsystem is None:
                    subsystem = self._subsystems[
                        subsystem_cls.NAME] = subsystem_cls(self)
                subsystem.add_grain(path, config)
        _logger.trace('Grain loading complete')

    def _try_get_config(self, path):
        yml = os.path.join(path, '.cob.yml')
        if os.path.isfile(yml):
            with open(yml) as f:
                return yaml.load(f.read())

        if os.path.isfile(path):
            with open(path) as f:
                return parse_front_matter(f)

        _logger.trace('No cob config detected ({}). Skipping...', path)
        return None

    def configure_app(self, flask_app):
        for subsystem in self:
            subsystem.activate(flask_app)

        for subsystem in self:
            subsystem.configure_app(flask_app)

    def _get_subsystem_by_grain_type(self, grain_type):
        return SubsystemBase.SUBSYSTEM_BY_NAME[grain_type]

    def __iter__(self):
        return iter(self._subsystems.values())


##########################################################################
# import all known subsystems to ensure registration
from . import flask_blueprint_subsystem  # pylint: disable=unused-import
from . import static_subsystem  # pylint: disable=unused-import
from . import models_subsystem  # pylint: disable=unused-import
from . import views_subsystem  # pylint: disable=unused-import
from . import templates_subsystem  # pylint: disable=unused-import
from . import frontend  # pylint: disable=unused-import



#!/usr/bin/env python
import sys

import logbook
import click

from ..bootstrapping import ensure_project_bootstrapped


@click.group()
@click.option("-v", "--verbose", count=True)
@click.option("-q", "--quiet", count=True)
def main(verbose, quiet):
    logbook.NullHandler().push_application()
    logbook.StreamHandler(sys.stderr, level=logbook.WARNING -
                          verbose + quiet, bubble=False).push_application()


@main.add_command
@click.command()
def bootstrap():
    ensure_project_bootstrapped()


def _add_all_subcommands():
    for name in [
            'develop',
            'docker',
            'generate',
            'testserver',
            'migrate',
            'db',
    ]:
        mod = __import__('cob.cli.{}'.format(name), fromlist=[''])
        cmd = getattr(mod, name)
        main.add_command(cmd)

_add_all_subcommands()


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



import logbook

from .base import SubsystemBase
from ..ctx import context

from flask_migrate import Migrate
from flask_sqlalchemy import SQLAlchemy


_logger = logbook.Logger(__name__)


class ModelsSubsystem(SubsystemBase):

    NAME = 'models'

    def activate(self, app):
        context.db = SQLAlchemy(app)
        Migrate(app, context.db).init_app(app)
        super(ModelsSubsystem, self).activate(app)

    def configure_grain(self, grain, app): # pylint: disable=unused-argument
        _logger.trace('Found models: {m.path}', grain)
        grain.load()



import os
import subprocess

_COB_DEVELOP_MODE = 'COB_DEVELOP'


def is_develop():
    return _COB_DEVELOP_MODE in os.environ


def cob_root():
    assert is_develop()
    return os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))


def cob_branch():
    git = subprocess.run(["git", "rev-parse", "--abbrev-ref", "HEAD"], cwd=cob_root(), stdout=subprocess.PIPE)
    git.check_returncode()
    return git.stdout.decode("UTF-8").strip()



import requests

import pytest


from .project import Project


def test_static_files_same_root():
    p = Project('staticfiles')
    with p.server_context() as url:
        assert requests.get(url.add_path('static1/file')).text == 'this is #1\n'
        assert requests.get(url.add_path('static1/file3')).text == 'this is #3\n'


def test_static_files():
    p = Project('staticfiles')
    with p.server_context() as url:
        assert requests.get(url.add_path('static1/file')).text == 'this is #1\n'
        assert requests.get(url.add_path('static2/file')).text == 'this is #2\n'



import click
import logbook

from ..app import build_app
from ..bootstrapping import ensure_project_bootstrapped

_logger = logbook.Logger(__name__)


@click.command()
@click.option('-p', '--port', type=int, default=5000)
def testserver(port):
    ensure_project_bootstrapped()
    flask_app = build_app()
    flask_app.run(port=port)



import os
import subprocess
from contextlib import contextmanager

import click
import jinja2
import logbook

import gossip

_logger = logbook.Logger(__name__)

_SKELETONS_ROOT = os.path.abspath(os.path.join(
    os.path.dirname(__file__), '..', '_skeletons'))

_ctx = {}

@click.group()
def generate():
    pass

@generate.command()
@click.option('-m', '--mountpoint', default=None)
@click.option('-t', '--type', default='views')
@click.argument('name')
def grain(type, name, mountpoint):
    if mountpoint is None:
        mountpoint = '/{}'.format(name)

    try:
        _generate('grain-{}'.format(type), name, {
            'name': name,
            'mountpoint': mountpoint,
        })
    except click.ClickException:
        raise click.ClickException('Unknown grain type {!r}'.format(type))

    gossip.trigger_with_tags('cob.after_generate.grain', tags=[type], kwargs={'name': name})

@gossip.register('cob.after_generate.grain', tags=['frontend-ember'])
def after_generate_grain_frontend_ember(*, name):
    status, _ = subprocess.getstatusoutput('which ember')
    if status != 0:
        click.echo('You do not seem to have ember-cli installed. Skipping ember app creation...')
    else:
        click.echo('Generating new Ember project')
        subprocess.check_call('ember init', cwd=name, shell=True)


@generate.command()
@click.argument('path')
@click.argument('project_name', required=False)
def project(path, project_name=None):
    if project_name is None:
        project_name = os.path.basename(path)
    _generate('project', path, {
        'project_name': project_name
    })

@generate.command()
@click.argument('name')
@click.argument('mountpoint', default='/static')
def static_dir(name, mountpoint):
    _generate('static_dir', name, {
        'name': name,
        'mountpoint': mountpoint,
    })
    os.mkdir(os.path.join(name, 'root'))

@generate.command()
@click.argument('name', default='models')
def models(name):
    _generate('models', name, {
        'name': name,
    })


def _generate(skeleton_name, dest_path, ctx):
    s = load_skeleton(skeleton_name)
    if s.is_single_file() and not dest_path.endswith('.py'):
        dest_path += '.py'
    with template_context(ctx):
        s.generate(dest_path)

@contextmanager
def template_context(ctx):
    _ctx.update(ctx)
    try:
        yield
    finally:
        _ctx.clear()


def load_skeleton(skeleton_name):
    skeleton_path = os.path.join(_SKELETONS_ROOT, skeleton_name)
    if not os.path.exists(skeleton_path):
        raise click.ClickException('No such skeleton: {!r}'.format(skeleton_name))

    if os.path.isdir(skeleton_path):
        return SkeletonDir(skeleton_path)

    return SkeletonFile(skeleton_path)


class Skeleton(object):

    def generate(self, dest_path):
        raise NotImplementedError()  # pragma: no cover

    def is_single_file(self):
        return isinstance(self, SkeletonFile)

class SkeletonDir(Skeleton):

    def __init__(self, path):
        super(SkeletonDir, self).__init__()
        self._path = path

    def generate(self, dest_path):
        if not os.path.isdir(dest_path):
            os.mkdir(dest_path)
        for filename in os.listdir(self._path):
            if filename == '.cob.git.keep':
                continue
            skeleton = load_skeleton(os.path.join(self._path, filename))
            skeleton.generate(os.path.join(dest_path, filename))


class SkeletonFile(Skeleton):

    def __init__(self, path):
        super(SkeletonFile, self).__init__()
        self._path = path

    def generate(self, dest_path):
        env = jinja2.Environment(loader=jinja2.FileSystemLoader(os.path.dirname(self._path)))
        template = env.get_template(os.path.basename(self._path))
        click.echo('Generating {}'.format(os.path.relpath(dest_path) if not os.path.isabs(dest_path) else dest_path))
        with open(dest_path, 'w') as f:
            f.write(template.render(**_ctx))



import os
from setuptools import setup, find_packages

with open(os.path.join(os.path.dirname(__file__), "cob", "__version__.py")) as version_file:
    exec(version_file.read()) # pylint: disable=W0122

_INSTALL_REQUIRES = [
    'click',
    'emport',
    'gossip',
    'Flask',
    'Flask-Migrate',
    'Flask-SQLAlchemy',
    'Jinja2',
    'gossip',
    'Logbook',
    'PyYAML',
    'tmuxp',
    'virtualenv',
]

setup(name="cob",
      classifiers = [
          "Programming Language :: Python :: 3.3",
          "Programming Language :: Python :: 3.4",
          "Programming Language :: Python :: 3.5",
          ],
      description="A scaffold for building and deploying powerful and robust webapps",
      license="BSD3",
      author="Rotem Yaari",
      author_email="vmalloc@gmail.com",
      version=__version__, # pylint: disable=E0602
      packages=find_packages(exclude=["tests"]),

      url="https://github.com/getcob/cob-cli",
      entry_points={
          "console_scripts": [
              "cob = cob.cli.main:main",
          ]},
      install_requires=_INSTALL_REQUIRES,
      scripts=[],
      namespace_packages=[],
      package_data={'cob': ['cob/Dockerfile.j2']},
      include_package_data=True,
      zip_safe=False,
      )



import re

import yaml


def parse_front_matter(fileobj):
    line = fileobj.readline().strip()

    match = re.match(r'^\s*#\s*cob:\s*(.*)$', line)
    if match:
        return _parse_oneline_dict(match.group(1))

    if re.match(r'^#\s?cob-yaml:\s*$', line):
        s = ""
        for line in fileobj:
            if not line.startswith('#'):
                return None

            line = line[1:]
            if line and line[0].isspace():
                line = line[1:]
            if line.strip() == '---':
                break
            s += line
        return yaml.load(s)

    return None

def _parse_oneline_dict(line):
    returned = {}
    for element in line.split():
        try:
            key, value = element.split('=')
        except ValueError:
            return None
        returned[key] = value
    return returned
